/*****************************************************************************
 *
 * FILE:	cockatrice.c
 * DESCRIPTION:	Cockatrice: simple CGI library
 * DATE:	Sat, Mar 16 2002
 * UPDATED:	Mon, Aug 20 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: cockatrice.c,v 1.3 2007/08/19 16:48:36 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#if	HAVE_STRING_H
#include <string.h>
#endif	/* HAVE_STRING_H */
#if	HAVE_DIRENT_H
#include <dirent.h>
#else
#define dirent	direct
#if	HAVE_SYS_NDIR_H
#include <sys/ndir.h>
#endif	/* HAVE_SYS_NDIR_H */
#if	HAVE_SYS_DIR_H
#include <sys/dir.h>
#endif	/* HAVE_SYS_DIR_H */
#if	HAVE_NDIR_H
#include <ndir.h>
#endif	/* HAVE_NDIR_H */
#endif	/* HAVE_DIRENT_H */
#include <sys/types.h>
#include <sys/mman.h>
#if     HAVE_SYS_TIME_H
#include <sys/time.h>
#endif  /* HAVE_SYS_TIME_H */
#include <sys/resource.h>
#include <signal.h>
#include <errno.h>
#include "cockatrice.h"
#include "module.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Global variables definition
 *
 *****************************************************************************/
int	cgi_errno;

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
#define	THRESHOLD_SIZE	(1024*2)

typedef struct {	/* query structure */
  char *	string;
  size_t	length;
  int		fd;	/* used by mmap() */
} Query;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void	set_sigtrap(void);
static void	set_limiter(void);
static void	cleanup(void);
static void	send_response_header(const char *);
static void	redirect_uri(const char *);
static int	read_data(void);
static int	parse_query(void);
static void	trap(int);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static CGI *	cgi = NULL;	/* cgi object */
static Query	query;		/* received data */

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	newCGI
 * DESCRIPTION:		create new CGI object
 * ARGUMENTS:		pluggable module list
 * RETURN VALUE:	CGI object
 */
CGI *
newCGI(modlist)
	int	modlist;
{
  if (cgi != NULL) {	/* called only once */
    cgi_errno = CGIERR_ALREADY_MADE_OBJECT;
    return NULL;
  }

  cgi = (CGI *)cc_calloc(1, sizeof(CGI));
  if (cgi != NULL) {
    char *	p;

    /* initialize query */
    query.string = NULL;
    query.length = 0;
    query.fd	 = -1;
    /**/
    set_sigtrap();
    set_limiter();
    /* initialize CGI entity */
    p = getenv("SCRIPT_NAME");
    if (p != NULL) {
      cgi->script_name = cc_strdup(p);
      if (cgi->script_name == NULL) {
	goto error;
      }
    }
    cgi->header	  = (void *)(send_response_header);
    cgi->redirect = (void *)(redirect_uri);
    /**/
    cgi->done	= (void *)(cleanup);
    /**/
    cgi->param	= (void *)(cc_param_get);
    cgi->params	= (void *)(cc_params_get);
    /* register static interfaces */
    cc_cookie_register_hooks();
    cc_upload_register_hooks();
    cc_utils_register_hooks();
    /* initialize */
    cc_param_init();
    cc_cookie_init();
    /* load modules */
    if (cc_load_modules(modlist) == -1) {
      goto error;
    }
    cc_initialize_modules();
    /* read data */
#if	DEBUG
    read_data();
#else
    if (read_data() == -1) {
      goto error;
    }
#endif	/* DEBUG */
    cgi_errno = CGIERR_NO_ERROR;

    return cgi;
  }

error:
  cc_free(cgi);

  return NULL;
}

/* set signal trap */
static void
set_sigtrap(void)
{
  struct sigaction	act;

  sigemptyset(&act.sa_mask);
  act.sa_flags	 = 0;
  act.sa_handler = (void *)trap;
  sigaction(SIGINT,  &act, NULL);
  sigaction(SIGTERM, &act, NULL);
  sigaction(SIGQUIT, &act, NULL);
  sigaction(SIGSEGV, &act, NULL);
  sigaction(SIGBUS,  &act, NULL);
  sigaction(SIGPIPE, &act, NULL);
}

/* limit resource */
static void
set_limiter(void)
{
  struct rlimit		rl;

  if (getrlimit(RLIMIT_CORE, &rl) == 0) {
    rl.rlim_cur = rl.rlim_max = 0;	/* no core dump */
    (void)setrlimit(RLIMIT_CORE, &rl);
  }
}

/*
 * FUNCTION NAME:	cleanup
 * DESCRIPTION:		cleanup all resources
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
cleanup(void)
{
  sigset_t	set;
  sigset_t	oset;

#if	0
  /* block all signals */
  sigfillset(&set);
  sigprocmask(SIG_BLOCK, &set, &oset);
#else
  sigprocmask(SIG_SETMASK, NULL, &set);
  sigprocmask(SIG_BLOCK, &set, &oset);
#endif

  if (query.string != NULL) {
    if (query.fd > 0) {
      munmap((char *)query.string, (size_t)query.length);
      close(query.fd);
      query.fd = -1;
    }
    else {
      free(query.string);
    }
    query.string = NULL;
  }

  cc_remove_modules();
  cc_upload_finish();
  cc_cookie_finish();
  cc_param_finish();

  cc_free(cgi->script_name);
  free(cgi);

  /* restore signal mask */
  sigprocmask(SIG_SETMASK, &oset, NULL);
}

/*
 * FUNCTION NAME:	send_response_header
 * DESCRIPTION:		send HTTP response header field
 * ARGUMENTS:		Content-Type
 * RETURN VALUE:	none
 */
static void
send_response_header(contentType)
	const char *	contentType;
{
  /* Content-Type response field */
  fputs("Content-Type: ", stdout);
  if (contentType != NULL) {
    fprintf(stdout, "%s" CRLF, contentType);
  }
  else {
    fputs("text/html" CRLF, stdout);
  }
  fputs(CRLF, stdout);
}

/*
 * FUNCTION NAME:	redirect_uri
 * DESCRIPTION:		flush out to Location response-header
 * ARGUMENTS:		URI
 * RETURN VALUE:	none
 */
static void
redirect_uri(uri)
	const char *	uri;
{
  fputs("Status: 301 Moved" CRLF, stdout);
  fprintf(stdout, "Location: %s" CRLF, uri);
  fprintf(stdout, "Content-Type: text/plain" CRLF);
  fputs(CRLF, stdout);
}

/*****************************************************************************/

/*
 * FUNCTION NAME:	read_data
 * DESCRIPTION:		read CGI data and set it to query buffer
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
read_data(void)
{
  char *	method;

  /* get REQUEST_METHOD */
  method = getenv("REQUEST_METHOD");
  if (method == NULL) {
    cgi_errno = CGIERR_NO_REQUEST_METHOD;
    return -1;
  }

  /* GET */
  if (strcasecmp(method, "GET") == 0) {
    char *	query_string = getenv("QUERY_STRING");

    if (query_string == NULL) {
      cgi_errno = CGIERR_NO_QUERY_STRING;
      return -1;
    }

    query.string = cc_strdup(query_string);
    if (query.string == NULL) {
      return -1;
    }
    query.length = strlen(query_string);
  }
  /* POST */
  else if (strcasecmp(method, "POST") == 0) {
    char *	contentLength	= getenv("CONTENT_LENGTH");
    char *	contentType	= getenv("CONTENT_TYPE");

    if (contentLength == NULL) {
      cgi_errno = CGIERR_NO_CONTENT_LENGTH;
      return -1;
    }
    query.length = (size_t)strtol(contentLength, (char **)NULL, 10);

    /* multipart form */
    if (contentType != NULL &&
	strstr(contentType, MULTIPART_FORM) != NULL) {
      return cc_upload_init();
    }
    else if (query.length < THRESHOLD_SIZE) {
      query.string = (char *)cc_calloc(query.length + 1, sizeof(char));
      if (query.string == NULL) {
	return -1;
      }
      if (fread(query.string, 1, query.length, stdin) != query.length) {
	cgi_errno = CGIERR_INTERNAL_LIBRARY_ERROR;
	return -1;
      }
    }
    else {
      char	tempfile[] = "/tmp/.cockatrice.XXXXXX";
      char	buf[BUFSIZ];
      size_t	len = 0;
      int	n;

      query.fd = mkstemp(tempfile);
      if (query.fd == -1) {
	cgi_errno = CGIERR_TMPFILE_OPEN_FAILURE;
	return -1;
      }
      unlink(tempfile);	/* the file is removed when it's closed. */

      memset(buf, 0, sizeof(buf));
      while ((n = read(0, buf, sizeof(buf))) > 0 && len < query.length) {
	write(query.fd, buf, (size_t)n);
	memset(buf, 0, sizeof(buf));
	len += n;
      }
      query.string = (char *)mmap(0,			/* address */
				 (size_t)query.length,	/* size */
				 PROT_READ|PROT_WRITE,	/* protection */
				 MAP_PRIVATE,	/* type of the mapped object */
				 query.fd,		/* file descriptor */
				 (off_t)0);		/* offset */
      if (len != query.length || query.string == MAP_FAILED) {
	close(query.fd);
	query.fd = -1;
	cgi_errno = CGIERR_INTERNAL_LIBRARY_ERROR;
	return -1;
      }
    }
  }
  /* error */
  else {
    cgi_errno = CGIERR_ILLEGAL_METHOD_TYPE;
    return -1;
  }

  /* parse */
  return parse_query();
}

/*
 * FUNCTION NAME:	parse_query
 * DESCRIPTION:		parse CGI query data
 * ARGUMENTS:		none
 * RETURN VALUE:	Successful: 0, otherwise: -1
 */
static int
parse_query(void)
{
  char *	str	= query.string;
  char *	tmp	= NULL;	/* top pointer for cc_uri_unescaped() */
  char *	name	= NULL;
  char *	value	= NULL;

  while ((str = strtok(str, "&")) != NULL) {
#if	ENABLE_ICONV || ENABLE_WKF
    char *	Kname;
    char *	Kvalue;
    size_t	len;
#endif	/* ENABLE_WKF */
    char *	p;

    name = tmp = cc_uri_unescape(str);
    p	 = (char *)strchr(name, '=');
    if (p != NULL) {
      *p++  = EOL;
      value = cc_chomp(p);
#if	ENABLE_ICONV || ENABLE_WKF
      len   = strlen(name) + 1;
      Kname = (char *)cc_calloc(len, sizeof(char));
      if (Kname != NULL && cc_conv_any2euc(name, len, Kname, len) == 0) {
	name = Kname;
      }
      len    = strlen(value) + 1;
      Kvalue = (char *)cc_calloc(len, sizeof(char));
      if (Kvalue != NULL && cc_conv_any2euc(value, len, Kvalue, len) == 0) {
	value = Kvalue;
      }
#endif	/* ENABLE_WKF */
      cc_param_put(name, value);
#if	ENABLE_ICONV || ENABLE_WKF
      cc_free(Kname);
      cc_free(Kvalue);
#endif	/* ENABLE_WKF */
      cc_free(tmp);
    }
    str = NULL;	/* DON'T REMOVE THIS LINE. */
  }

  return 0;
}

/*****************************************************************************/

void
cc_hook_response_header(routine)
	void 	(*routine)(const char *);
{
  cgi->header = routine;
}

void
cc_hook_redirect_uri(routine)
	void 	(*routine)(const char *);
{
  cgi->redirect = routine;
}

void
cc_hook_module_uri(routine)
	const cmURI *	routine;
{
  cgi->uri = routine;
}

void
cc_hook_module_file(routine)
	const cmFile *	routine;
{
  cgi->file = routine;
}

void
cc_hook_module_util(routine)
	const cmUtil *	routine;
{
  cgi->util = routine;
}

void
cc_hook_module_cookie(routine)
	const cmCookie *	routine;
{
  cgi->cookie = routine;
}

void
cc_hook_module_upload(routine)
	const cmUpload *	routine;
{
  cgi->upload = routine;
}

void
cc_hook_module_env(routine)
	const cmEnv *	routine;
{
  cgi->env = routine;
}

void
cc_hook_module_session(routine)
	const cmSession *	routine;
{
  cgi->session = routine;
}

void
cc_hook_module_database(routine)
	const cmDatabase *	routine;
{
  cgi->database = routine;
}

void
cc_hook_module_date(routine)
	const cmDate *	routine;
{
  cgi->date = routine;
}

void
cc_hook_module_check(routine)
	const cmCheck *	routine;
{
  cgi->check = routine;
}

/*****************************************************************************/

/*
 * Function Name:	trap
 * Description:		trap all signals
 * Arguments:		signal
 * Return Value:	never return
 */
static void
trap(sig)
	int	sig;
{
  fprintf(stderr, "[notice] cockatrice: terminated by sig#%d\n", sig);
  cleanup();
  exit(1);
}
