/*****************************************************************************
 *
 * FILE:	mod_env.c
 * DESCRIPTION:	Cockatrice: environment variable manipulator (module)
 * DATE:	Sun, Mar 17 2002
 * UPDATED:	Mon, Aug 20 2007
 * AUTHOR:	Kouichi ABE (WALL) / 
 * E-MAIL:	kouichi@MysticWALL.COM
 * URL:		http://www.MysticWALL.COM/
 * COPYRIGHT:	(c) 2002-2007 졿Kouichi ABE (WALL), All rights reserved.
 * LICENSE:
 *
 *  Copyright (c) 2002-2007 Kouichi ABE (WALL) <kouichi@MysticWALL.COM>,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *   1. Redistributions of source code must retain the above copyright
 *      notice, this list of conditions and the following disclaimer.
 *
 *   2. Redistributions in binary form must reproduce the above copyright
 *      notice, this list of conditions and the following disclaimer in the
 *      documentation and/or other materials provided with the distribution.
 *
 *   THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 *   ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 *   THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 *   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE
 *   LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 *   CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 *   SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 *   INTERRUPTION)  HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 *   CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 *   ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 *   THE POSSIBILITY OF SUCH DAMAGE.
 *
 * $Id: mod_env.c,v 1.1 2007/08/19 16:49:15 kouichi Exp $
 *
 *****************************************************************************/

#if	HAVE_CONFIG_H
#include "config.h"
#endif	/* HAVE_CONFIG_H */

#include <stdio.h>
#if	HAVE_STDLIB_H
#include <stdlib.h>
#endif	/* HAVE_STDLIB_H */
#if	HAVE_UNISTD_H
#include <unistd.h>
#endif	/* HAVE_UNISTD_H */
#include "cockatrice.h"
#include "module.h"
#include "functions.h"

/*****************************************************************************
 *
 *	Macros and structures definition
 *
 *****************************************************************************/
typedef struct {
  char *	server_software;
  char *	server_name;
  char *	gateway_interface;
  /**/
  char *	server_protocol;
  int		server_port;
  /**/
  char *	request_method;
  char *	script_name;
  char *	path_info;
  char *	path_translated;
  char *	remote_host;
  char *	remote_addr;
  int		remote_port;
  char *	auth_type;
  char *	remote_user;
  char *	remote_ident;
  /**/
  char *	accept;
  char *	accept_encoding;
  char *	accept_language;
  char *	user_agent;
  char *	referer;
  /**/
  char *	document_root;
  char *	server_root;
  char *	server_admin;
} Env;

/*****************************************************************************
 *
 *	Local functions declaration
 *
 *****************************************************************************/
static void		cm_env_init(void);
static void		cm_env_finish(void);
static char *		cm_env_get(const char *);
static const char *	cm_env_server_software(void);
static const char *	cm_env_server_name(void);
static const char *	cm_env_gateway_interface(void);
static const char *	cm_env_server_protocol(void);
static int		cm_env_server_port(void);
static const char *	cm_env_request_method(void);
static const char *	cm_env_script_name(void);
static const char *	cm_env_path_info(void);
static const char *	cm_env_path_translated(void);
static const char *	cm_env_remote_host(void);
static const char *	cm_env_remote_addr(void);
static int		cm_env_remote_port(void);
static const char *	cm_env_auth_type(void);
static const char *	cm_env_remote_user(void);
static const char *	cm_env_remote_ident(void);
static const char *	cm_env_http_accept(void);
static const char *	cm_env_http_accept_encoding(void);
static const char *	cm_env_http_accept_language(void);
static const char *	cm_env_http_user_agent(void);
static const char *	cm_env_http_referer(void);
static const char *	cm_env_document_root(void);
static const char *	cm_env_server_root(void);
static const char *	cm_env_server_admin(void);

/*****************************************************************************
 *
 *	Local variables declaration
 *
 *****************************************************************************/
static Env *	env;

/*****************************************************************************
 *
 *	Functions definition
 *
 *****************************************************************************/

/*
 * FUNCTION NAME:	cm_env_init
 * DESCRIPTION:		initialize environment variable
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
cm_env_init(void)
{
  env = (Env *)cc_malloc(sizeof(Env));
  if (env != NULL) {
    register char *	p;	/* working pointer */

    env->server_software= getenv("SERVER_SOFTWARE");
    env->server_name	= getenv("SERVER_NAME");
    env->gateway_interface= getenv("GATEWAY_INTERFACE");
    /**/
    env->server_protocol= getenv("SERVER_PROTOCOL");

    p			= getenv("SERVER_PORT");
    env->server_port	= p == NULL ? -1 : (int)strtol(p, (char **)NULL, 10);
    /**/
    env->request_method	= getenv("REQUEST_METHOD");
    env->script_name	= getenv("SCRIPT_NAME");
    env->path_info	= getenv("PATH_INFO");
    env->path_translated= getenv("PATH_TRANSLATED");
    env->remote_host	= getenv("REMOTE_HOST");
    env->remote_addr	= getenv("REMOTE_ADDR");

    p			= getenv("REMOTE_PORT");
    env->remote_port	= p == NULL ? -1 : (int)strtol(p, (char **)NULL, 10);
  
    env->auth_type	= getenv("AUTH_TYPE");
    env->remote_user	= getenv("REMOTE_USER");
    env->remote_ident	= getenv("REMOTE_IDENT");
    /**/
    env->accept		= getenv("HTTP_ACCEPT");
    env->accept_encoding= getenv("HTTP_ACCEPT_ENCODING");
    env->accept_language= getenv("HTTP_ACCEPT_LANGUAGE");
    env->user_agent	= getenv("HTTP_USER_AGENT");
    env->referer	= getenv("HTTP_REFERER");
    /**/
    env->document_root	= getenv("DOCUMENT_ROOT");
    env->server_root	= getenv("SERVER_ROOT");
    env->server_admin	= getenv("SERVER_ADMIN");
  }
}

/*
 * FUNCTION NAME:	cm_env_finish
 * DESCRIPTION:		release memory area
 * ARGUMENTS:		none
 * RETURN VALUE:	none
 */
static void
cm_env_finish(void)
{
  free(env);
}

/*****************************************************************************/

static char *
cm_env_get(name)
	const char *	name;
{
  return getenv(name);
}

static const char *
cm_env_server_software(void)
{
  return env->server_software;
}

static const char *
cm_env_server_name(void)
{
  return env->server_name;
}

static const char *
cm_env_gateway_interface(void)
{
  return env->gateway_interface;
}

/**/

static const char *
cm_env_server_protocol(void)
{
  return env->server_protocol;
}

static int
cm_env_server_port(void)
{
  return env->server_port;
}

/**/

static const char *
cm_env_request_method(void)
{
  return env->request_method;
}

static const char *
cm_env_script_name(void)
{
  return env->script_name;
}

static const char *
cm_env_path_info(void)
{
  return env->path_info;
}

static const char *
cm_env_path_translated(void)
{
  return env->path_translated;
}

static const char *
cm_env_remote_host(void)
{
  return env->remote_host;
}

static const char *
cm_env_remote_addr(void)
{
  return env->remote_addr;
}

static int
cm_env_remote_port(void)
{
  return env->remote_port;
}

static const char *
cm_env_auth_type(void)
{
  return env->auth_type;
}

static const char *
cm_env_remote_user(void)
{
  return env->remote_user;
}

static const char *
cm_env_remote_ident(void)
{
  return env->remote_ident;
}

/**/

static const char *
cm_env_http_accept(void)
{
  return env->accept;
}

static const char *
cm_env_http_accept_encoding(void)
{
  return env->accept_encoding;
}

static const char *
cm_env_http_accept_language(void)
{
  return env->accept_language;
}

static const char *
cm_env_http_user_agent(void)
{
  return env->user_agent;
}

static const char *
cm_env_http_referer(void)
{
  return env->referer;
}

/**/

static const char *
cm_env_document_root(void)
{
  return env->document_root;
}

static const char *
cm_env_server_root(void)
{
  return env->server_root;
}

static const char *
cm_env_server_admin(void)
{
  return env->server_admin;
}

/*****************************************************************************/

static const cmEnv	env_funcs = {
  cm_env_get,
  cm_env_server_software,
  cm_env_server_name,
  cm_env_gateway_interface,
  cm_env_server_protocol,
  cm_env_server_port,
  cm_env_request_method,
  cm_env_script_name,
  cm_env_path_info,
  cm_env_path_translated,
  cm_env_remote_host,
  cm_env_remote_addr,
  cm_env_remote_port,
  cm_env_auth_type,
  cm_env_remote_user,
  cm_env_remote_ident,
  cm_env_http_accept,
  cm_env_http_accept_encoding,
  cm_env_http_accept_language,
  cm_env_http_user_agent,
  cm_env_http_referer,
  cm_env_document_root,
  cm_env_server_root,
  cm_env_server_admin
};

static void
cm_env_register_hooks(void)
{
  cc_hook_module_env(&env_funcs);
}

static ccModule	mod_env = {
  CC_MODULE_MAGIC_NUMBER_MAJOR,		/* API version */
  "mod_env",				/* module name */
  "Cockatrice: environment module 1.0",	/* description such as version etc.. */
  NULL,					/* module handle */
  (void *)cm_env_init,			/* called when it's loaded */
  (void *)cm_env_finish,		/* called when it's unloaded */
  (void *)cm_env_register_hooks
};

ccModule *
ccModuleInfo(void)
{
  return &mod_env;
}
